function [ joint, marginals, prior, map, dims, coherent ] = findJoint_KL(Marginals, Graphs, laplace_const)

if(isempty(Marginals))
    joint = [];
    marginals = [];
    map = [];
    coherent = [];
    prior = [];
    dims = [];
    return;
end

% In our case 4: -->, <--, <-->, -/-
nevents = length(Marginals{1}.w);

if(numel(Marginals) == 1)
    joint = Marginals{1}.w';
    marginals = Marginals{1}.w;
    map = 1:nevents;
    coherent = true;
    prior = computeUninformativePrior(Marginals, Graphs, laplace_const);
    dims = nevents;
    return;
end

% Initialize variables
npk = length(Marginals);
nworlds = nevents^npk;
dims = zeros(1,npk) + nevents;

% Compute prior distribution
prior = computeUninformativePrior(Marginals, Graphs, laplace_const);

% Create linear constraint matrices
valid = getValidConfs(Marginals, 1:length(Marginals)) + 1;
nvalid = length(valid);
a = zeros(nevents*npk,nvalid);

idx_c = nevents*((1:npk)-1);

confs = mdInd2Sub_mex(dims,valid);
% a_idx = (confs + repmat(idx_c,nvalid,1))';
a_idx = (confs' + reshape(squeeze(idx_c(:,:,ones(nvalid,1))),size(confs,2),size(confs,1)));

base = (0:nvalid-1)' * nevents * npk;
a(a_idx + base(:,ones(npk,1))') = 1;

% Alternatively
% a(a_idx + repmat(base', npk, 1)) = 1;

% for i = 1:nvalid
% %     cur = idx_c + confs(i,:);
% %     curp = mdInd2Sub_mex(dims,valid(i));
%     a(a_idx(:,i),i) = 1;
% %     a_idx(:,i) = cur;
% end


% Constructing the map
map = zeros(nworlds,1);
map(valid) = find(valid);
if(npk > 1)
    map = reshape(map,zeros(1,npk)+nevents);
end

marginals = getMarginals(Marginals);

% Check for coherence
options = optimset('Diagnostics','off','Display','off','LargeScale','off','Algorithm','simplex');
[~,~,exitflag] = linprog([],[],[],a,marginals,zeros(size(a,2),1),ones(size(a,2),1),[],options);
coherent = exitflag > 0;

% Run step Make-Coherent for each bi-connected component!

a_logical = a ~= 0;
joint = prior;

iter = 0;
while(true)
%     iter
    p_next = joint;
    
    if(coherent)
        h_ = marginals;
    else
        % Computed weighted projection
        h_ratio = marginals./(a * joint);
        p_next_ = 1/npk * joint .* sum(h_ratio(a_idx))';
        % Compute marginals
        h_ = a * p_next_;
    end
    
    % I-projection of Q(n) to Q(n+1)
    %     h_n = a * p;
    %     h_ratio = h_ ./ h_n;
    %     h_ratio_p = h_ratio .^ (1/npk);
    %     p_next = p .* prod(h_ratio_p(a_idx))';
    
    %     for i = 1:length(marginals)
    %         h_n = a(i,:) * p_next;
    %         h_ratio = h_(i)./h_n;
    %         p_next = p_next .* (h_ratio .^ a(i,:))';
    %     end
    
    for i = 1:length(h_)
        cur_idx = a_logical(i,:);
        p_next(cur_idx) = p_next(cur_idx) .* (h_(i)./sum(p_next(cur_idx)));
    end
    
%     max(abs(joint - p_next))
    if(all(abs(joint - p_next) <= 10^-6) || iter >= 10000)
        marginals = a * joint;
%         fprintf('Finished after %d iterations\n',iter);
        break;
    end
    
    joint = p_next;
    iter = iter + 1;
end

end

function theta = getMarginals(Marginals)
nevents = length(Marginals{1}.w);
theta = zeros(nevents*length(Marginals),1);
for i = 1:length(Marginals)
    theta(nevents*(i-1)+1:nevents*(i-1)+nevents) = Marginals{i}.w;
end
end
