function [Components, initial_idx] = getComponents(Marginals, connected)
% connected: true for connected components, false for bi-connected
% components

% Initialize variables
vmap = mapVertices(Marginals);
nnodes = length(vmap);
npk = length(Marginals);

% Initializing variables for the prior knowledge
st = zeros(npk,2);
for idx = 1:npk
    st(idx,1) = vmap{Marginals{idx}.s};
    st(idx,2) = vmap{Marginals{idx}.t};
end

% Create undirected graph with prior knowledge
G = zeros(nnodes,nnodes);
for i = 1:npk
    G(st(i,1),st(i,2)) = 1;
    G(st(i,2),st(i,1)) = 1;
end

if(connected)
    % Find connected components
    [Ci,~] = components(sparse(G));
    C = G;
    for i = 1:length(Ci)
        C(i,C(i,:) > 0) = Ci(i);
    end
else
    % Find bi-connected components
    [~,C] = biconnected_components(sparse(G));
end

% Split PK to independent Components
ncomponents = max(C(:));

Components = cell(ncomponents,1);

for i = 1:ncomponents
    Components{i} = [];
end

initial_idx = [];
for i = 1:npk
    Components{C(st(i,1),st(i,2))}(end+1) = i;
end

for i = 1:ncomponents
   initial_idx = [initial_idx Components{i}]; 
end

[~,initial_idx] = sort(initial_idx);
end
