function [valid, partialvalid] = getValidConfs(Marginals, marginal_indices)

if(isempty(Marginals))
    valid = [];
    return;
end

npk = length(Marginals);
power_indices_partial = marginal_indices-1;
power_indices = 0:npk-1;

% Split to bi-connected components
Components = getComponents(Marginals, false);
ncomponents = length(Components);

Valid = cell(ncomponents,1);
PartialValid = cell(ncomponents,1);

% For each component, find all valid configurations
component_sizes = zeros(ncomponents,1);
for i = 1:ncomponents
    component_sizes(i) = length(Components{i});
end

for i = 1:ncomponents
    [Valid{i}, PartialValid{i}]= innerFunction(Marginals(Components{i}), power_indices(Components{i}), power_indices_partial(Components{i}));
end

valid = sum(allcomb(Valid{:}),2);
partialvalid = sum(allcomb(PartialValid{:}),2);

[valid,i] = sort(valid);
partialvalid = partialvalid(i);
end

function [valid, partialvalid] = innerFunction(Marginals, power_indices, power_indices_partial)

% In our case 4: -->, <--, <-->, -/-
nevents = length(Marginals{1}.w);

% Initialize variables
[vmap, nnodes] = mapVertices(Marginals);
npk = length(Marginals);
nworlds = nevents^npk;

% Initializing variables for the prior knowledge
st_mapped = zeros(npk,2);
for idx = 1:npk
    st_mapped(idx,1) = vmap{Marginals{idx}.s};
    st_mapped(idx,2) = vmap{Marginals{idx}.t};
end

% Create all event combinations to avoid costly mdInd2Sub calls
a = cell(npk,1);
for i = 1:npk
    a{i} = 1:nevents;
end
c = allcomb(a{:});

valid = false(nworlds,1);
for i = 1:nworlds
    valid(i) = confIsValid(st_mapped, nnodes, c(i,:));
end

partialvalid = (c(valid,:)-1) * nevents.^power_indices_partial';
valid = (c(valid,:)-1) * nevents.^power_indices';

end
