#include "mex.h"

#if MX_API_VER < 0x07030000
typedef int mwIndex;
typedef int mwSize;
#endif /* MX_API_VER */

#include <math.h>
#include <stdlib.h>
#include <string.h>

void innerFunction(const double *dims, const double *idx, mwSize ndims, mwSize nidx, double *result);

void mexFunction(int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[])
{
    mwSize ndims, nidx;
    
    /* output data */
    double *result;
    
    /* arguments */
    const mxArray* dims;
    const mxArray* idx;
    
    if (nrhs != 2) {
        mexErrMsgTxt("2 inputs required.");
    }
    
    dims = prhs[0];
    idx = prhs[1];
    
    ndims = max(mxGetN(dims), mxGetM(dims));
    nidx = max(mxGetN(idx), mxGetM(idx));
    
    plhs[0] = mxCreateDoubleMatrix(nidx,ndims,mxREAL);
    result = mxGetPr(plhs[0]);
    
    innerFunction(mxGetPr(dims), mxGetPr(idx), ndims, nidx, result);
}

void innerFunction(const double *dims, const double *idx, mwSize ndims, mwSize nidx, double *result) {
    unsigned int curidx;
    unsigned int i,j;
    
    for(j = 0; j < nidx; ++j) {
        curidx = (int) idx[j] - 1;
        for(i = 0; i < ndims; ++i) {
            *(result + i*nidx + j) = (curidx % (int) dims[i]) + 1;
            curidx = (int) (curidx/dims[i]);
        }
    }
}
