function [fullprior,Priors] = computeUninformativePrior(Marginals, Graphs, laplace_const)

% Split to independent components
Components = getComponents(Marginals, true);
ncomponents = length(Components);
% npk = length(Marginals);
nevents = length(Marginals{1}.w);

% For each components, find valid configurations
Valid = cell(ncomponents,1);
PartialValid = cell(ncomponents,1);

for i = 1:ncomponents
    [Valid{i}, PartialValid{i}] = getValidConfs(Marginals(Components{i}), Components{i});
end

% For each component, compute prior
Priors = cell(ncomponents,1);
for i = 1:ncomponents
    Priors{i} = computePrior(Marginals(Components{i}), Graphs, Valid{i}+1, nevents, 0);
end

% Compute joint
fullprior = prod(allcomb(Priors{:}),2);
ngraphs = length(Graphs);
fullprior = (fullprior * ngraphs + laplace_const)/(ngraphs + length(fullprior)*laplace_const);

% Change the order to match the initially given order
order = sum(allcomb(PartialValid{:}),2);

[~,i] = sort(order,'ascend');
fullprior = fullprior(i);

end
