function [G, score, graphscore, priorscore] = searchAndScore(data, domain_counts, joint, G0, N_prime, swapOperator, screen)

    function closure_ = updateClosure(closure_,i,j)
        closure_(i,j) = 1;
        X = closure_(:,i);
        X(i) = 1;
        Y = closure_(j,:);
        Y(j) = 1;
        closure_ = closure_ + (X*Y);
    end

    function updateScores_Graph(G,newG)
        changes = false(1,nnodes);
        for n = 1:nnodes
            changes(n) = any(any(G(:,n) ~= newG(:,n)));
        end
        
        % Update Local Scores
        for n = 1:nnodes
            if(changes(n))
                localscores(n) = bdeu(data,domain_counts,n,find(newG(:,n)'),N_prime);
            end
        end
        
        % Update Insert/Delete Scores
        for n = 1:nnodes
            if(changes(n))
                updateScores_Node(newG,n);
            end
        end
        updateReverse(newG);
        updatePKScores(newG);
    end

    function updateScores_Node(G,n2)
        % Update all insert entries for n2
        for n1 = 1:nnodes
            if(n1 == n2)
                continue;
            end
            
            if(G(n1,n2) == 0)
                G_ = G; G_(n1,n2) = 1;
                insert(n1,n2,1) = bdeu(data,domain_counts,n2,find(G_(:,n2))',N_prime) - localscores(n2);
                remove(n1,n2,1) = 0;
            else
                G_ = G; G_(n1,n2) = 0;
                remove(n1,n2,1) = bdeu(data,domain_counts,n2,find(G_(:,n2))',N_prime) - localscores(n2);
                insert(n1,n2,1) = 0;
            end
        end
    end

    function updateReverse(G)
        [I,J] = find(G ~= 0);
        nedges = length(I);
        for e = 1:nedges
            n1 = I(e); n2 = J(e);
            reverse(n1,n2,1) = remove(n1,n2,1) + insert(n2,n1,1);
        end
        
        [I,J] = find(G == 0);
        nedges = length(I);
        for e = 1:nedges
            n1 = I(e); n2 = J(e);
            reverse(n1,n2,1) = 0;
        end
    end

    function updatePKScores(G)
        [I,J] = find(~G');
        nedges = length(I);
        
        for e = 1:nedges
            n1 = I(e); n2 = J(e);
            if(n1 == n2)
                continue;
            end
            
            % Insert case
            conf_ = conf;
            change = false;
            
            % Pre-caching
            cond = closure(:,n1);
            cond(n1) = 1;
            
            for c = 1:npk
                s = st(c,1);
                t = st(c,2);
                
                % Pre-caching
                ancestor_n2_s = (n2 == s || closure(n2,s));
                ancestor_n2_t = (n2 == t || closure(n2,t));
                
                if(xor(ancestor_n2_s,ancestor_n2_t))
                    if(conf(c) == 3)
                        if(ancestor_n2_t && (s == n1 || closure(s,n1)))
                            conf_(c) = 1;
                            change = true;
                        elseif(ancestor_n2_s && (t == n1 || closure(t,n1)))
                            conf_(c) = 2;
                            change = true;
                        end
                    elseif(conf(c) == 4)
                        if(ancestor_n2_t && (s == n1 || closure(s,n1)))
                            conf_(c) = 1;
                            change = true;
                        elseif(ancestor_n2_s && (t == n1 || closure(t,n1)))
                            conf_(c) = 2;
                            change = true;
                            % If exists W s.t. W-->s && W-->n1, n2-->s || W-->t && W-->n1, n2 -->t
                        elseif((ancestor_n2_t && any(closure(:,s) & cond)) || (ancestor_n2_s && any(closure(:,t) & cond)))
                            conf_(c) = 3;
                            change = true;
                        end
                    end
                end
            end
            if(change)
                insert(n1,n2,2) = PKScore(conf_,joint) - priorscore;
            else
                insert(n1,n2,2) = 0;
            end
        end
        
        [I,J] = find(G');
        nedges = length(I);
        for e = 1:nedges
            insert(I(e),J(e),2) = 0;
        end
        
        for n1 = 1:nnodes
            for n2 = 1:nnodes
                if(n1 == n2)
                    continue;
                end
                
                % Delete case
                if(G(n1,n2) == 1)
                    G_ = G; G_(n1,n2) = 0;
                    closure_ = transitiveClosure_mex(G_);
                    conf_ = getPKConf(closure_,st);
                    
                    if(any(conf ~= conf_))
                        remove(n1,n2,2) = PKScore(conf_,joint) - priorscore;
                    else
                        remove(n1,n2,2) = 0;
                    end
                else
                    remove(n1,n2,2) = 0;
                end
                
                % Reverse case (also check if no other path from n1 to n2?)
                if(G(n1,n2) == 1)
                    G_ = G; G_(n1,n2) = 0; G_(n2,n1) = 1;
                    closure_ = transitiveClosure_mex(G_);
                    if(~closure_(n1,n2))
                        conf_ = getPKConf(closure_,st);
                        if(any(conf ~= conf_))
                            reverse(n1,n2,2) = PKScore(conf_,joint) - priorscore;
                        else
                            reverse(n1,n2,2) = 0;
                        end
                    else
                        reverse(n1,n2,2) = 0;
                    end
                else
                    reverse(n1,n2,2) = 0;
                end
            end
        end
    end

    function operation = getMaxOperation
        operation.scorediff = 0;
        operation.op = [];
        operation.variables = [];
        
        % Add edge
        [I,J] = find(~G & ~closure');
        nedges = length(I);
        for e = 1:nedges
            n1 = I(e); n2 = J(e);
            curdiff = insert(n1,n2,1) + insert(n1,n2,2);
            if(curdiff > operation.scorediff)
                operation.scorediff = curdiff;
                operation.op = 1;
                operation.variables = [n1 n2];
            end
        end
        
        % Remove edge
        [I,J] = find(G);
        nedges = length(I);
        for e = 1:nedges
            n1 = I(e); n2 = J(e);
            curdiff = remove(n1,n2,1) + remove(n1,n2,2);
            if(curdiff > operation.scorediff)
                operation.scorediff = curdiff;
                operation.op = 2;
                operation.variables = [n1 n2];
            end
        end
        
        % Reverse edge
        [I,J] = find(G);
        nedges = length(I);
        for e = 1:nedges
            n1 = I(e); n2 = J(e);
            G_ = G; G_(n1,n2) = 0; G_(n2,n1) = 1;
            
            if(~isReachableSparse_mex(sparse(G_),n1,n2))
                curdiff = reverse(n1,n2,1) + reverse(n1,n2,2);
                if(curdiff > operation.scorediff)
                    operation.scorediff = curdiff;
                    operation.op = 3;
                    operation.variables = [n1 n2];
                end
            end
        end
    end

nnodes = size(G0,1);
threshold = 10^-8;
G = G0;
closure = transitiveClosureSparse_mex(sparse(G));

if(isempty(joint))
    st = [];
    priorscore = 0;
else
    npk = 0;
    for i = 1:length(joint)
        npk = npk + length(joint{i}.marginal_idx);
    end
    st = zeros(npk,2);
    for i = 1:length(joint)
       st(joint{i}.marginal_idx,:) = joint{i}.st; 
    end
    conf = getPKConf(closure, st);
    priorscore = PKScore(conf,joint);
end

% Initialize local scores
localscores = zeros(1,nnodes);
for i = 1:nnodes
    localscores(i) = bdeu(data,domain_counts,i,find(G(:,i)'),N_prime);
end

% Initialize total scores
graphscore = sum(localscores);

% Initialize score and valid matrices
insert = zeros(nnodes,nnodes,2);
remove = zeros(nnodes,nnodes,2);
reverse = zeros(nnodes,nnodes,2);

for i = 1:nnodes
    updateScores_Node(G,i);
end
if(~isempty(joint))
    updatePKScores(G);
end
updateReverse(G);

while(true)
    operation = getMaxOperation();
    
    if(operation.scorediff <= threshold)
        break;
    end
    
    if(operation.op == 1) % Add edge
        i = operation.variables(1);
        j = operation.variables(2);
        localscores(j) = insert(i,j,1) + localscores(j);
        priorscore = priorscore + insert(i,j,2);
        G(i,j) = 1;
        closure = updateClosure(closure,i,j);
        conf = getPKConf(closure, st);
        updateScores_Node(G,j);
        if(screen)
            fprintf('Adding %d->%d, diff: %f',i,j,operation.scorediff);
        end
    elseif(operation.op == 2) % Remove edge
        i = operation.variables(1);
        j = operation.variables(2);
        localscores(j) = remove(i,j,1) + localscores(j);
        priorscore = priorscore + remove(i,j,2);
        G(i,j) = 0;
        closure = transitiveClosureSparse_mex(sparse(G));
        conf = getPKConf(closure, st);
        updateScores_Node(G,j);
        if(screen)
            fprintf('Removing %d->%d, diff: %f',i,j,operation.scorediff);
        end
    elseif(operation.op == 3) % Reverse edge
        i = operation.variables(1);
        j = operation.variables(2);
        localscores(i) = insert(j,i,1) + localscores(i);
        localscores(j) = remove(i,j,1) + localscores(j);
        priorscore = priorscore + reverse(i,j,2);
        G(i,j) = 0;
        G(j,i) = 1;
        closure = transitiveClosureSparse_mex(sparse(G));
        conf = getPKConf(closure, st);
        updateScores_Node(G,i);
        updateScores_Node(G,j);
        if(screen)
            fprintf('Reversing %d->%d, diff: %f',i,j,operation.scorediff);
        end
    else
        error('Unknown operator');
    end
    
    if(~isempty(joint))
        updatePKScores(G);
    end
    updateReverse(G);

    graphscore = sum(localscores);
    
    if(~isempty(joint) && swapOperator)
        PDAG = convertDagToPdag(G,false);
        uncertainties = sum(sum(PDAG.G == 1))/2;
        
        if(uncertainties > 0)
            [newG,newpriorscore] = incorporatePK_PDAG_BnB(PDAG, joint, G, priorscore);
            
            if(newpriorscore > priorscore)
                if(screen)
                    fprintf('\n\t*** Swapping %d, diff: %f',sum(sum(G ~= newG))/2,newpriorscore-priorscore);
                end
                closure = transitiveClosureSparse_mex(sparse(newG));
                conf = getPKConf(closure,st);
                priorscore = newpriorscore;
                updateScores_Graph(G,newG);
                G = newG;
            end
        end
    end
    if(screen)
        fprintf('\n');
    end
end

score = graphscore + priorscore;

end
