function [G,score] = incorporatePK_PDAG_BnB(PDAG, joint, G_init, score_init)

global nnodes;
global curscore;
global curgraph;
global indices;
global ncomp;
global nevents;
global nworlds;
global npk;
global st;

if(isempty(joint))
    G = [];
    score = [];
    return;
end

nevents = 4;
ncomp = length(joint);
npk = zeros(ncomp,1);
for i = 1:ncomp
   npk(i) = length(joint{i}.marginal_idx);
end
st = zeros(sum(npk),2);
for i = 1:length(joint)
    st(joint{i}.marginal_idx,:) = joint{i}.st;
end
nworlds = nevents.^npk;

nnodes = length(PDAG.G);
curgraph = G_init;
curscore = score_init;
indices = createIndices(joint);

pdag = PDAG.G;
closure = transitiveClosurePag_mex(pdag);

% Partialconf: cell array, ncomp elements, each element size(npk,nevents)
% idx: cell array, ncomp elements, each element size(nevents^npk)

partialconf_old = cell(ncomp,1);
for i = 1:ncomp
   partialconf_old{i} = true(npk(i), nevents); 
end
idx = cell(ncomp,1);
for i = 1:ncomp
   idx{i} = true(nworlds(i),1); 
end

partialconf = updatePartialConf(pdag, joint, closure, partialconf_old);
idx = updateIdx(idx, partialconf_old, partialconf);
search_all(pdag, joint, partialconf, G_init, MaxPKScore(joint, idx), idx);

G = curgraph;
score = curscore;
end

function indices = createIndices(Jk)

global ncomp;
global nworlds;

% Optimize: pre-compute valid indices, pass as argument to
% incorporatePK_PDAG_BnB

indices = cell(ncomp,1);
for i = 1:ncomp
   indices{i} = mdInd2Sub_mex(Jk{i}.dims, 1:nworlds(i));
end

end

function idx = updateIdx(idx, partialconf_old, partialconf_new)
global ncomp
global indices;
global npk;

for c = 1:ncomp
    cur_old = partialconf_old{c};
    cur_new = partialconf_new{c};
    curindices = indices{c};
    curidx = idx{c};
    for i = 1:npk(c)
        if(cur_old(i,1) ~= cur_new(i,1))
            curidx = curidx & curindices(:,i) ~= 1;
        end
        if(cur_old(i,2) ~= cur_new(i,2))
            curidx = curidx & curindices(:,i) ~= 2;
        end
        if(cur_old(i,3) ~= cur_new(i,3))
            curidx = curidx & curindices(:,i) ~= 3;
        end
        if(cur_old(i,4) ~= cur_new(i,4))
            curidx = curidx & curindices(:,i) ~= 4;
        end
    end
    idx{c} = curidx;
end

end

function [partialconf, change] = updatePartialConf(pdag, Jk, closure, partialconf)

global ncomp;
global npk;

change = false;

potentialclosure = transitiveClosureSparse_mex(sparse((pdag == 2 & pdag' == 3) | (pdag == 1)));
for c = 1:ncomp
    curpartialconf = partialconf{c};
    curst = Jk{c}.st;
    for i = 1:npk(c)
        s = sum(curpartialconf(i,:));
        if(s > 1) % No definite configuration found
            if(closure(curst(i,1),curst(i,2)))
                curpartialconf(i,1) = true;
                curpartialconf(i,[2 3 4]) = false;
                change = true;
            elseif(closure(curst(i,2),curst(i,1)))
                curpartialconf(i,2) = true;
                curpartialconf(i,[1 3 4]) = false;
                change = true;
            elseif(any(closure(:,curst(i,1)) & closure(:,curst(i,2))))
                curpartialconf(i,3) = true;
                curpartialconf(i,[1 2 4]) = false;
                change = true;
            else
                curpartialconf(i,1) = potentialclosure(curst(i,1),curst(i,2));
                curpartialconf(i,2) = potentialclosure(curst(i,2),curst(i,1));
                curpartialconf(i,3) = any(potentialclosure(:,curst(i,1)) & potentialclosure(:,curst(i,2)));
                
                if(sum(curpartialconf(i,:)) ~= s)
                    change = true;
                end
            end
        end
    end
    partialconf{c} = curpartialconf;
end

end

function score = MaxPKScore(Jk, idx)
global ncomp;

score = 0;

for i = 1:ncomp
    midx = Jk{i}.map(idx{i});
    midx = midx(midx > 0);
    score = score + max(Jk{i}.score(midx));
end

end

function sat = search_all(pdag, Jk, partialconf, G_init, maxscore, idx)
global curscore;
global curgraph;
global st;

if(maxscore <= curscore)
    sat = false;
    return;
end

[X,Y] = find(pdag == 1,1);
% Edge heuristic here + dont use known values
% [X,Y] = find(bn == 1,1);

if(isempty(X))
    G = double(pdag == 2 & pdag' == 3);
    closure = transitiveClosureSparse_mex(sparse(G));
    assert(~hasCycle(closure));
    
    conf = getPKConf(closure,st);
    score = PKScore(conf,Jk);
    
    if(score > curscore)
        curscore = score;
        curgraph = G;
    elseif(score == curscore && ~isempty(G_init))
        % If same score, choose the one closer to the original graph
        distance1 = sum(sum(G == G_init));
        distance2 = sum(sum(curgraph == G_init));
        if(distance1 > distance2)
            curgraph = G;
        end
    end
    sat = true;
    return;
end

% Test arrow
pdag1 = pdag;
pdag1(X,Y) = 2;
pdag1(Y,X) = 3;
pdag1 = PC_rules_apply(pdag1, 0);
closure1 = transitiveClosurePag_mex(pdag1);
[partialconf1, change1] = updatePartialConf(pdag1, Jk, closure1, partialconf);

if(change1)
    idx1 = updateIdx(idx, partialconf, partialconf1);
    sat1 = search_all(pdag1,Jk,partialconf1,G_init, MaxPKScore(Jk, idx1), idx1);
else
    sat1 = search_all(pdag1,Jk,partialconf1,G_init, maxscore, idx);
end

assert(~hasCycle(closure1));

% Test tail
pdag2 = pdag;
pdag2(X,Y) = 3;
pdag2(Y,X) = 2;
pdag2 = PC_rules_apply(pdag2, 0);
closure2 = transitiveClosurePag_mex(pdag2);
[partialconf2, change2] = updatePartialConf(pdag2, Jk, closure2, partialconf);

if(change2)
    idx2 = updateIdx(idx, partialconf, partialconf2);
    sat2 = search_all(pdag2,Jk,partialconf2,G_init, MaxPKScore(Jk, idx2), idx2);
else
    sat2 = search_all(pdag2,Jk,partialconf2,G_init, maxscore, idx);
end

assert(~hasCycle(closure2));

sat = sat1 || sat2;

end
