%% 1) Load data, graph, domain_counts
% data: MxN matrix, with M samples and N variables
% graph: NxN adjacency matrix
% domain_counts: the number of values for each variable (e.g. 2 for a
% binary variable)

load alarm.mat;
[nsamples, nnodes] = size(data);

%% 2) Sampling DAGs
fprintf('Sampling Graphs...\n');
ngraphsamples = 1000;
[Graphs, a_n] = sampleDags(ngraphsamples, nnodes);
for i = 1:ngraphsamples
    Graphs{i} = transitiveClosure_mex(Graphs{i});
end
% singleprior contains the uninformative prior distribution of a single
% path variable
singleprior = computeSinglePrior(Graphs);


% Suggestion: For a given number of nodes you just have to sample graphs
% once. Thus, in case you want to repeat an experiment for a given dataset
% you just have to sample the graphs once. I suggest to sample a
% sufficiently large number of graphs and save them, including the values
% for a_n and singleprior.

%% 3) Compute Joint Distribution from Prior Knowledge
fprintf('Computing Joint Distribution...\n');
laplace_const = 1;

% Example of creating prior knowledge from a known graph, with probability
% = 0.9 for the real relation between two variables.

% P{1} contains the marginal beliefs for the real relation between
% variables 1 and 2 with probability 0.9, and the remaining probabilities
% distributed in an uninformative way. P{1} is a struct with fields:
% .s: index first variable
% .t: index second variable
% .w: [P(s-->t), P(s<--t), P(s<-->t), P(s-/-t)]

% See the code of createPK_weighted in case you have prior knowledge 
% (e.g. P(x-->y) = p) and want to set the remaining probabilities in an
% uninformative way.
P{1} = createPK_weighted(graph,[1,2],singleprior,0.9);
P{2} = createPK_weighted(graph,[1,3],singleprior,0.9);
P{3} = createPK_weighted(graph,[1,4],singleprior,0.9);
P{4} = createPK_weighted(graph,[2,3],singleprior,0.9);
P{5} = createPK_weighted(graph,[4,5],singleprior,0.9);


% 'full': the full exhaustive joint distribution (<= 10 beliefs)
% 'partition': the approximate method (<= 10 beliefs for each connected component)
% 'independent': assume all beliefs are independent (any number of beliefs)

% Joint = computeJoint(P, Graphs, laplace_const, a_n, 'full', true);
Joint = computeJoint(P, Graphs, laplace_const, a_n, 'partition', true);
% Joint = computeJoint(P, Graphs, laplace_const, a_n, 'independent', true);

%% 4) Search-and-Score

% Starting graph (empty in this case)
G0 = zeros(nnodes);

% Equivalent sample size for the bdeu score
N_prime = 1;

% Use the swap equivalent operator or not
useSwapOperator = true;

% Print steps to screen
screen = true;

DAG = searchAndScore(data, domain_counts, Joint, G0, N_prime, useSwapOperator, screen);